'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var React = require('react');

function isLens(thing) {
    return thing instanceof BaseLens;
}

function observableSymbol() {
    return (typeof Symbol === "function" && Symbol.observable) || "@@observable";
}

var lensId = 0;
var BaseLens = /** @class */ (function () {
    function BaseLens() {
        this.selectorCache = new Map();
        // optimization: initialize fields as empty
        // optimization: no forEach
        this.derivations = [];
        this.parents = [];
        this.lensId = ++lensId;
        this.dirty = 0;
        this.changedParents = 0;
        this.hot = false;
        this.computed = false;
    }
    BaseLens.prototype.propagateChanged = function () {
        if (++this.dirty === 1) {
            this.changedDerivations = this.derivations.slice();
            this.changedDerivations.forEach(function (d) { return d.propagateChanged(); });
        }
    };
    BaseLens.prototype.propagateReady = function (changed) {
        if (changed)
            this.changedParents++;
        if (this.dirty > 0 && --this.dirty === 0) {
            if (this.changedParents) {
                this.changedParents = 0;
                var old = this.state;
                this.state = this.recompute();
                this.computed = true;
                if (this.state !== old) {
                    this.changedDerivations.forEach(function (d) {
                        return d.propagateReady(true);
                    });
                    return;
                }
            }
            this.changedDerivations.forEach(function (d) { return d.propagateReady(false); });
            this.changedDerivations = undefined;
        }
    };
    BaseLens.prototype.value = function () {
        var res = !this.hot
            ? this.recompute()
            : this.computed
                ? this.state
                : (this.computed = true, this.state = this.recompute());
        notifyRead(this);
        return res;
    };
    BaseLens.prototype.subscribe = function (handler) {
        return this.do(subscribe(handler));
    };
    BaseLens.prototype[observableSymbol()] = function () {
        return this.do(toStream);
    };
    BaseLens.prototype.registerDerivation = function (lens) {
        var resume = this.derivations.length === 0;
        this.derivations.push(lens);
        if (resume) {
            this.hot = true;
        }
        var cacheKey = lens.getCacheKey();
        if (cacheKey !== undefined) {
            // not entirely happy on only caching selectors as long as the lens is hot,
            // but without a map of weak references we can otherwise not prevent leaking memory...
            this.selectorCache.set(cacheKey, lens);
        }
        if (resume) {
            this.resume();
        }
    };
    BaseLens.prototype.removeDerivation = function (lens) {
        var idx = this.derivations.indexOf(lens);
        if (idx === -1)
            fail$1("Illegal state"); // todo fail
        this.derivations.splice(idx, 1);
        var cacheKey = lens.getCacheKey();
        if (cacheKey !== undefined)
            this.selectorCache.delete(cacheKey);
        if (this.derivations.length === 0) {
            this.hot = false;
            this.computed = false;
            this.state = undefined; // prevent leaking mem
            this.suspend();
        }
    };
    BaseLens.prototype.do = function () {
        var things = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            things[_i] = arguments[_i];
        }
        return things.reduce(function (acc, transformer, index) {
            if (!isLens(acc))
                return fail$1("The transformer at argument " + (index -
                    1) + " of '.do()' should produce a lens, but it returned a: (" + typeof acc + ") '" + acc);
            // So, this is ok even if acc is not a lens anymore, just becomes a general chain...
            if (typeof transformer !== "function")
                return fail$1("Not a valid transformer at argument " + index + " of '.do()', got: (" + typeof transformer + ") '" + transformer + "'");
            return transformer(acc);
        }, this);
    };
    BaseLens.prototype.toString = function () {
        return "Lens[" + this.describe() + "\n]";
    };
    BaseLens.prototype.transform = function (config) {
        if (config.cacheKey !== undefined &&
            this.selectorCache.has(config.cacheKey))
            return this.selectorCache.get(config.cacheKey);
        return new Pipe(this, config);
    };
    BaseLens.prototype.select = function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return this.do.apply(this, args.map(select));
    };
    return BaseLens;
}());

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = Object.setPrototypeOf ||
    ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
    function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var __assign = Object.assign || function __assign(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
    }
    return t;
};

var Pipe = /** @class */ (function (_super) {
    __extends(Pipe, _super);
    function Pipe(base, config) {
        var _this = _super.call(this) || this;
        _this.base = base;
        _this.nextUpdate = function (updater) {
            _this.base.update(updater);
        };
        var _a = config.cacheKey, cacheKey = _a === void 0 ? undefined : _a, _b = config.onNext, onNext = _b === void 0 ? defaultRecompute : _b, _c = config.onUpdate, onUpdate = _c === void 0 ? defaultUpdate : _c, _d = config.description, description = _d === void 0 ? "(unknown pipe)" : _d, _e = config.onSuspend, onSuspend = _e === void 0 ? noop : _e, _f = config.onResume, onResume = _f === void 0 ? noop : _f;
        _this.config = {
            cacheKey: cacheKey,
            onNext: onNext,
            onUpdate: onUpdate,
            description: description,
            onSuspend: onSuspend,
            onResume: onResume
        };
        return _this;
    }
    Pipe.prototype.recompute = function () {
        return this.config.onNext(this.base.value(), this.state, this);
    };
    Pipe.prototype.update = function (updater) {
        this.config.onUpdate(normalizeUpdater(updater), this.nextUpdate, this);
    };
    Pipe.prototype.getCacheKey = function () {
        return this.config.cacheKey;
    };
    Pipe.prototype.resume = function () {
        
        this.base.registerDerivation(this);
        this.config.onResume();
    };
    Pipe.prototype.suspend = function () {
        
        this.base.removeDerivation(this);
        this.config.onSuspend();
    };
    Pipe.prototype.describe = function () {
        return (this.base.describe() +
            "\n\t.do(" +
            this.config.description +
            ")");
    };
    return Pipe;
}(BaseLens));
function defaultRecompute(nextValue) {
    return nextValue;
}
function defaultUpdate(updater, next) {
    next(updater); // note: eliminate some function in Pipe.update to make stack friendlier?
}

function generatePatches(state, basepath, patches, inversePatches, baseValue, resultValue) {
    if (patches) if (Array.isArray(baseValue)) generateArrayPatches(state, basepath, patches, inversePatches, baseValue, resultValue);else generateObjectPatches(state, basepath, patches, inversePatches, baseValue, resultValue);
}

function generateArrayPatches(state, basepath, patches, inversePatches, baseValue, resultValue) {
    var shared = Math.min(baseValue.length, resultValue.length);
    for (var i = 0; i < shared; i++) {
        if (state.assigned[i] && baseValue[i] !== resultValue[i]) {
            var path = basepath.concat(i);
            patches.push({ op: "replace", path: path, value: resultValue[i] });
            inversePatches.push({ op: "replace", path: path, value: baseValue[i] });
        }
    }
    if (shared < resultValue.length) {
        // stuff was added
        for (var _i = shared; _i < resultValue.length; _i++) {
            var _path = basepath.concat(_i);
            patches.push({ op: "add", path: _path, value: resultValue[_i] });
        }
        inversePatches.push({
            op: "replace",
            path: basepath.concat("length"),
            value: baseValue.length
        });
    } else if (shared < baseValue.length) {
        // stuff was removed
        patches.push({
            op: "replace",
            path: basepath.concat("length"),
            value: resultValue.length
        });
        for (var _i2 = shared; _i2 < baseValue.length; _i2++) {
            var _path2 = basepath.concat(_i2);
            inversePatches.push({ op: "add", path: _path2, value: baseValue[_i2] });
        }
    }
}

function generateObjectPatches(state, basepath, patches, inversePatches, baseValue, resultValue) {
    each(state.assigned, function (key, assignedValue) {
        var origValue = baseValue[key];
        var value = resultValue[key];
        var op = !assignedValue ? "remove" : key in baseValue ? "replace" : "add";
        if (origValue === baseValue && op === "replace") return;
        var path = basepath.concat(key);
        patches.push(op === "remove" ? { op: op, path: path } : { op: op, path: path, value: value });
        inversePatches.push(op === "add" ? { op: "remove", path: path } : op === "remove" ? { op: "add", path: path, value: origValue } : { op: "replace", path: path, value: origValue });
    });
}

function applyPatches(draft, patches) {
    var _loop = function _loop(i) {
        var patch = patches[i];
        if (patch.path.length === 0 && patch.op === "replace") {
            draft = patch.value;
        } else {
            var path = patch.path.slice();
            var key = path.pop();
            var base = path.reduce(function (current, part) {
                if (!current) throw new Error("Cannot apply patch, path doesn't resolve: " + patch.path.join("/"));
                return current[part];
            }, draft);
            if (!base) throw new Error("Cannot apply patch, path doesn't resolve: " + patch.path.join("/"));
            switch (patch.op) {
                case "replace":
                case "add":
                    // TODO: add support is not extensive, it does not support insertion or `-` atm!
                    base[key] = patch.value;
                    break;
                case "remove":
                    if (Array.isArray(base)) {
                        if (key === base.length - 1) base.length -= 1;else throw new Error("Remove can only remove the last key of an array, index: " + key + ", length: " + base.length);
                    } else delete base[key];
                    break;
                default:
                    throw new Error("Unsupported patch operation: " + patch.op);
            }
        }
    };

    for (var i = 0; i < patches.length; i++) {
        _loop(i);
    }
    return draft;
}

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
};



















var defineProperty = function (obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
};

var NOTHING = typeof Symbol !== "undefined" ? Symbol("immer-nothing") : defineProperty({}, "immer-nothing", true);

var PROXY_STATE = typeof Symbol !== "undefined" ? Symbol("immer-proxy-state") : "__$immer_state";

var RETURNED_AND_MODIFIED_ERROR = "An immer producer returned a new value *and* modified its draft. Either return a new value *or* modify the draft.";

function verifyMinified() {}

var inProduction = typeof process !== "undefined" && process.env.NODE_ENV === "production" || verifyMinified.name !== "verifyMinified";

var autoFreeze = !inProduction;
var useProxies = typeof Proxy !== "undefined";

function getUseProxies() {
    return useProxies;
}

function isProxy(value) {
    return !!value && !!value[PROXY_STATE];
}

function isProxyable(value) {
    if (!value) return false;
    if ((typeof value === "undefined" ? "undefined" : _typeof(value)) !== "object") return false;
    if (Array.isArray(value)) return true;
    var proto = Object.getPrototypeOf(value);
    return proto === null || proto === Object.prototype;
}

function freeze(value) {
    if (autoFreeze) {
        Object.freeze(value);
    }
    return value;
}

var assign = Object.assign || function assign(target, value) {
    for (var key in value) {
        if (has(value, key)) {
            target[key] = value[key];
        }
    }
    return target;
};

function shallowCopy(value) {
    if (Array.isArray(value)) return value.slice();
    var target = value.__proto__ === undefined ? Object.create(null) : {};
    return assign(target, value);
}

function each(value, cb) {
    if (Array.isArray(value)) {
        for (var i = 0; i < value.length; i++) {
            cb(i, value[i]);
        }
    } else {
        for (var key in value) {
            cb(key, value[key]);
        }
    }
}

function has(thing, prop) {
    return Object.prototype.hasOwnProperty.call(thing, prop);
}

// given a base object, returns it if unmodified, or return the changed cloned if modified
function finalize(base, path, patches, inversePatches) {
    if (isProxy(base)) {
        var state = base[PROXY_STATE];
        if (state.modified === true) {
            if (state.finalized === true) return state.copy;
            state.finalized = true;
            var result = finalizeObject(useProxies ? state.copy : state.copy = shallowCopy(base), state, path, patches, inversePatches);
            generatePatches(state, path, patches, inversePatches, state.base, result);
            return result;
        } else {
            return state.base;
        }
    }
    finalizeNonProxiedObject(base);
    return base;
}

function finalizeObject(copy, state, path, patches, inversePatches) {
    var base = state.base;
    each(copy, function (prop, value) {
        if (value !== base[prop]) {
            // if there was an assignment on this property, we don't need to generate
            // patches for the subtree
            var _generatePatches = patches && !has(state.assigned, prop);
            copy[prop] = finalize(value, _generatePatches && path.concat(prop), _generatePatches && patches, inversePatches);
        }
    });
    return freeze(copy);
}

function finalizeNonProxiedObject(parent) {
    // If finalize is called on an object that was not a proxy, it means that it is an object that was not there in the original
    // tree and it could contain proxies at arbitrarily places. Let's find and finalize them as well
    if (!isProxyable(parent)) return;
    if (Object.isFrozen(parent)) return;
    each(parent, function (i, child) {
        if (isProxy(child)) {
            parent[i] = finalize(child);
        } else finalizeNonProxiedObject(child);
    });
    // always freeze completely new data
    freeze(parent);
}



function is(x, y) {
    // From: https://github.com/facebook/fbjs/blob/c69904a511b900266935168223063dd8772dfc40/packages/fbjs/src/core/shallowEqual.js
    if (x === y) {
        return x !== 0 || 1 / x === 1 / y;
    } else {
        return x !== x && y !== y;
    }
}

// @ts-check

var proxies = null;

var objectTraps = {
    get: get$1,
    has: function has$$1(target, prop) {
        return prop in source(target);
    },
    ownKeys: function ownKeys(target) {
        return Reflect.ownKeys(source(target));
    },

    set: set$1,
    deleteProperty: deleteProperty,
    getOwnPropertyDescriptor: getOwnPropertyDescriptor,
    defineProperty: defineProperty$1,
    setPrototypeOf: function setPrototypeOf() {
        throw new Error("Immer does not support `setPrototypeOf()`.");
    }
};

var arrayTraps = {};
each(objectTraps, function (key, fn) {
    arrayTraps[key] = function () {
        arguments[0] = arguments[0][0];
        return fn.apply(this, arguments);
    };
});
arrayTraps.deleteProperty = function (state, prop) {
    if (isNaN(parseInt(prop))) throw new Error("Immer does not support deleting properties from arrays: " + prop);
    return objectTraps.deleteProperty.call(this, state[0], prop);
};
arrayTraps.set = function (state, prop, value) {
    if (prop !== "length" && isNaN(parseInt(prop))) throw new Error("Immer does not support setting non-numeric properties on arrays: " + prop);
    return objectTraps.set.call(this, state[0], prop, value);
};

function createState(parent, base) {
    return {
        modified: false, // this tree is modified (either this object or one of it's children)
        assigned: {}, // true: value was assigned to these props, false: was removed
        finalized: false,
        parent: parent,
        base: base,
        copy: undefined,
        proxies: {}
    };
}

function source(state) {
    return state.modified === true ? state.copy : state.base;
}

function get$1(state, prop) {
    if (prop === PROXY_STATE) return state;
    if (state.modified) {
        var value = state.copy[prop];
        if (value === state.base[prop] && isProxyable(value))
            // only create proxy if it is not yet a proxy, and not a new object
            // (new objects don't need proxying, they will be processed in finalize anyway)
            return state.copy[prop] = createProxy(state, value);
        return value;
    } else {
        if (has(state.proxies, prop)) return state.proxies[prop];
        var _value = state.base[prop];
        if (!isProxy(_value) && isProxyable(_value)) return state.proxies[prop] = createProxy(state, _value);
        return _value;
    }
}

function set$1(state, prop, value) {
    // TODO: optimize
    state.assigned[prop] = true;
    if (!state.modified) {
        if (prop in state.base && is(state.base[prop], value) || has(state.proxies, prop) && state.proxies[prop] === value) return true;
        markChanged(state);
    }
    state.copy[prop] = value;
    return true;
}

function deleteProperty(state, prop) {
    state.assigned[prop] = false;
    markChanged(state);
    delete state.copy[prop];
    return true;
}

function getOwnPropertyDescriptor(state, prop) {
    var owner = state.modified ? state.copy : has(state.proxies, prop) ? state.proxies : state.base;
    var descriptor = Reflect.getOwnPropertyDescriptor(owner, prop);
    if (descriptor && !(Array.isArray(owner) && prop === "length")) descriptor.configurable = true;
    return descriptor;
}

function defineProperty$1() {
    throw new Error("Immer does not support defining properties on draft objects.");
}

function markChanged(state) {
    if (!state.modified) {
        state.modified = true;
        state.copy = shallowCopy(state.base);
        // copy the proxies over the base-copy
        Object.assign(state.copy, state.proxies); // yup that works for arrays as well
        if (state.parent) markChanged(state.parent);
    }
}

// creates a proxy for plain objects / arrays
function createProxy(parentState, base, key) {
    if (isProxy(base)) throw new Error("Immer bug. Plz report.");
    var state = createState(parentState, base, key);
    var proxy = Array.isArray(base) ? Proxy.revocable([state], arrayTraps) : Proxy.revocable(state, objectTraps);
    proxies.push(proxy);
    return proxy.proxy;
}

function produceProxy(baseState, producer, patchListener) {
    if (isProxy(baseState)) {
        // See #100, don't nest producers
        var returnValue = producer.call(baseState, baseState);
        return returnValue === undefined ? baseState : returnValue;
    }
    var previousProxies = proxies;
    proxies = [];
    var patches = patchListener && [];
    var inversePatches = patchListener && [];
    try {
        // create proxy for root
        var rootProxy = createProxy(undefined, baseState);
        // execute the thunk
        var _returnValue = producer.call(rootProxy, rootProxy);
        // and finalize the modified proxy
        var result = void 0;
        // check whether the draft was modified and/or a value was returned
        if (_returnValue !== undefined && _returnValue !== rootProxy) {
            // something was returned, and it wasn't the proxy itself
            if (rootProxy[PROXY_STATE].modified) throw new Error(RETURNED_AND_MODIFIED_ERROR);

            // See #117
            // Should we just throw when returning a proxy which is not the root, but a subset of the original state?
            // Looks like a wrongly modeled reducer
            result = finalize(_returnValue);
            if (patches) {
                patches.push({ op: "replace", path: [], value: result });
                inversePatches.push({ op: "replace", path: [], value: baseState });
            }
        } else {
            result = finalize(rootProxy, [], patches, inversePatches);
        }
        // revoke all proxies
        each(proxies, function (_, p) {
            return p.revoke();
        });
        patchListener && patchListener(patches, inversePatches);
        return result;
    } finally {
        proxies = previousProxies;
    }
}

// @ts-check

var descriptors = {};
var states = null;

function createState$1(parent, proxy, base) {
    return {
        modified: false,
        assigned: {}, // true: value was assigned to these props, false: was removed
        hasCopy: false,
        parent: parent,
        base: base,
        proxy: proxy,
        copy: undefined,
        finished: false,
        finalizing: false,
        finalized: false
    };
}

function source$1(state) {
    return state.hasCopy ? state.copy : state.base;
}

function _get(state, prop) {
    assertUnfinished(state);
    var value = source$1(state)[prop];
    if (!state.finalizing && value === state.base[prop] && isProxyable(value)) {
        // only create a proxy if the value is proxyable, and the value was in the base state
        // if it wasn't in the base state, the object is already modified and we will process it in finalize
        prepareCopy(state);
        return state.copy[prop] = createProxy$1(state, value);
    }
    return value;
}

function _set(state, prop, value) {
    assertUnfinished(state);
    state.assigned[prop] = true; // optimization; skip this if there is no listener
    if (!state.modified) {
        if (is(source$1(state)[prop], value)) return;
        markChanged$1(state);
        prepareCopy(state);
    }
    state.copy[prop] = value;
}

function markChanged$1(state) {
    if (!state.modified) {
        state.modified = true;
        if (state.parent) markChanged$1(state.parent);
    }
}

function prepareCopy(state) {
    if (state.hasCopy) return;
    state.hasCopy = true;
    state.copy = shallowCopy(state.base);
}

// creates a proxy for plain objects / arrays
function createProxy$1(parent, base) {
    var proxy = shallowCopy(base);
    each(base, function (i) {
        Object.defineProperty(proxy, "" + i, createPropertyProxy("" + i));
    });
    var state = createState$1(parent, proxy, base);
    createHiddenProperty(proxy, PROXY_STATE, state);
    states.push(state);
    return proxy;
}

function createPropertyProxy(prop) {
    return descriptors[prop] || (descriptors[prop] = {
        configurable: true,
        enumerable: true,
        get: function get$$1() {
            return _get(this[PROXY_STATE], prop);
        },
        set: function set$$1(value) {
            _set(this[PROXY_STATE], prop, value);
        }
    });
}

function assertUnfinished(state) {
    if (state.finished === true) throw new Error("Cannot use a proxy that has been revoked. Did you pass an object from inside an immer function to an async process? " + JSON.stringify(state.copy || state.base));
}

// this sounds very expensive, but actually it is not that expensive in practice
// as it will only visit proxies, and only do key-based change detection for objects for
// which it is not already know that they are changed (that is, only object for which no known key was changed)
function markChangesSweep() {
    // intentionally we process the proxies in reverse order;
    // ideally we start by processing leafs in the tree, because if a child has changed, we don't have to check the parent anymore
    // reverse order of proxy creation approximates this
    for (var i = states.length - 1; i >= 0; i--) {
        var state = states[i];
        if (state.modified === false) {
            if (Array.isArray(state.base)) {
                if (hasArrayChanges(state)) markChanged$1(state);
            } else if (hasObjectChanges(state)) markChanged$1(state);
        }
    }
}

function markChangesRecursively(object) {
    if (!object || (typeof object === "undefined" ? "undefined" : _typeof(object)) !== "object") return;
    var state = object[PROXY_STATE];
    if (!state) return;
    var proxy = state.proxy,
        base = state.base;

    if (Array.isArray(object)) {
        if (hasArrayChanges(state)) {
            markChanged$1(state);
            state.assigned.length = true;
            if (proxy.length < base.length) for (var i = proxy.length; i < base.length; i++) {
                state.assigned[i] = false;
            } else for (var _i = base.length; _i < proxy.length; _i++) {
                state.assigned[_i] = true;
            }each(proxy, function (index, child) {
                if (!state.assigned[index]) markChangesRecursively(child);
            });
        }
    } else {
        var _diffKeys = diffKeys(base, proxy),
            added = _diffKeys.added,
            removed = _diffKeys.removed;

        if (added.length > 0 || removed.length > 0) markChanged$1(state);
        each(added, function (_, key) {
            state.assigned[key] = true;
        });
        each(removed, function (_, key) {
            state.assigned[key] = false;
        });
        each(proxy, function (key, child) {
            if (!state.assigned[key]) markChangesRecursively(child);
        });
    }
}

function diffKeys(from, to) {
    // TODO: optimize
    var a = Object.keys(from);
    var b = Object.keys(to);
    return {
        added: b.filter(function (key) {
            return a.indexOf(key) === -1;
        }),
        removed: a.filter(function (key) {
            return b.indexOf(key) === -1;
        })
    };
}

function hasObjectChanges(state) {
    var baseKeys = Object.keys(state.base);
    var keys = Object.keys(state.proxy);
    return !shallowEqual(baseKeys, keys);
}

function hasArrayChanges(state) {
    var proxy = state.proxy;

    if (proxy.length !== state.base.length) return true;
    // See #116
    // If we first shorten the length, our array interceptors will be removed.
    // If after that new items are added, result in the same original length,
    // those last items will have no intercepting property.
    // So if there is no own descriptor on the last position, we know that items were removed and added
    // N.B.: splice, unshift, etc only shift values around, but not prop descriptors, so we only have to check
    // the last one
    var descriptor = Object.getOwnPropertyDescriptor(proxy, proxy.length - 1);
    // descriptor can be null, but only for newly created sparse arrays, eg. new Array(10)
    if (descriptor && !descriptor.get) return true;
    // For all other cases, we don't have to compare, as they would have been picked up by the index setters
    return false;
}

function produceEs5(baseState, producer, patchListener) {
    if (isProxy(baseState)) {
        // See #100, don't nest producers
        var returnValue = producer.call(baseState, baseState);
        return returnValue === undefined ? baseState : returnValue;
    }
    var prevStates = states;
    states = [];
    var patches = patchListener && [];
    var inversePatches = patchListener && [];
    try {
        // create proxy for root
        var rootProxy = createProxy$1(undefined, baseState);
        // execute the thunk
        var _returnValue = producer.call(rootProxy, rootProxy);
        // and finalize the modified proxy
        each(states, function (_, state) {
            state.finalizing = true;
        });
        var result = void 0;
        // check whether the draft was modified and/or a value was returned
        if (_returnValue !== undefined && _returnValue !== rootProxy) {
            // something was returned, and it wasn't the proxy itself
            if (rootProxy[PROXY_STATE].modified) throw new Error(RETURNED_AND_MODIFIED_ERROR);
            result = finalize(_returnValue);
            if (patches) {
                patches.push({ op: "replace", path: [], value: result });
                inversePatches.push({ op: "replace", path: [], value: baseState });
            }
        } else {
            if (patchListener) markChangesRecursively(rootProxy);
            markChangesSweep(); // this one is more efficient if we don't need to know which attributes have changed
            result = finalize(rootProxy, [], patches, inversePatches);
        }
        // make sure all proxies become unusable
        each(states, function (_, state) {
            state.finished = true;
        });
        patchListener && patchListener(patches, inversePatches);
        return result;
    } finally {
        states = prevStates;
    }
}

function shallowEqual(objA, objB) {
    //From: https://github.com/facebook/fbjs/blob/c69904a511b900266935168223063dd8772dfc40/packages/fbjs/src/core/shallowEqual.js
    if (is(objA, objB)) return true;
    if ((typeof objA === "undefined" ? "undefined" : _typeof(objA)) !== "object" || objA === null || (typeof objB === "undefined" ? "undefined" : _typeof(objB)) !== "object" || objB === null) {
        return false;
    }
    var keysA = Object.keys(objA);
    var keysB = Object.keys(objB);
    if (keysA.length !== keysB.length) return false;
    for (var i = 0; i < keysA.length; i++) {
        if (!hasOwnProperty.call(objB, keysA[i]) || !is(objA[keysA[i]], objB[keysA[i]])) {
            return false;
        }
    }
    return true;
}

function createHiddenProperty(target, prop, value) {
    Object.defineProperty(target, prop, {
        value: value,
        enumerable: false,
        writable: true
    });
}

/**
 * produce takes a state, and runs a function against it.
 * That function can freely mutate the state, as it will create copies-on-write.
 * This means that the original state will stay unchanged, and once the function finishes, the modified state is returned
 *
 * @export
 * @param {any} baseState - the state to start with
 * @param {Function} producer - function that receives a proxy of the base state as first argument and which can be freely modified
 * @param {Function} patchListener - optional function that will be called with all the patches produced here
 * @returns {any} a new state, or the base state if nothing was modified
 */
function produce(baseState, producer, patchListener) {
    // prettier-ignore
    if (arguments.length < 1 || arguments.length > 3) throw new Error("produce expects 1 to 3 arguments, got " + arguments.length);

    // curried invocation
    if (typeof baseState === "function") {
        // prettier-ignore
        if (typeof producer === "function") throw new Error("if first argument is a function (curried invocation), the second argument to produce cannot be a function");

        var initialState = producer;
        var recipe = baseState;

        return function () {
            var args = arguments;

            var currentState = args[0] === undefined && initialState !== undefined ? initialState : args[0];

            return produce(currentState, function (draft) {
                args[0] = draft; // blegh!
                return recipe.apply(draft, args);
            });
        };
    }

    // prettier-ignore
    {
        if (typeof producer !== "function") throw new Error("if first argument is not a function, the second argument to produce should be a function");
        if (patchListener !== undefined && typeof patchListener !== "function") throw new Error("the third argument of a producer should not be set or a function");
    }

    // if state is a primitive, don't bother proxying at all
    if ((typeof baseState === "undefined" ? "undefined" : _typeof(baseState)) !== "object" || baseState === null) {
        var returnValue = producer(baseState);
        return returnValue === undefined ? baseState : normalizeResult(returnValue);
    }

    if (!isProxyable(baseState)) throw new Error("the first argument to an immer producer should be a primitive, plain object or array, got " + (typeof baseState === "undefined" ? "undefined" : _typeof(baseState)) + ": \"" + baseState + "\"");
    return normalizeResult(getUseProxies() ? produceProxy(baseState, producer, patchListener) : produceEs5(baseState, producer, patchListener));
}

function normalizeResult(result) {
    return result === NOTHING ? undefined : result;
}

var applyPatches$1 = produce(applyPatches);

var nothing = NOTHING;

var storeId = 0;
var Store = /** @class */ (function (_super) {
    __extends(Store, _super);
    function Store(initialValue, options) {
        var _this = _super.call(this) || this;
        _this.isRunningUpdate = false;
        _this.name = options && options.name ? options.name : "Store" + ++storeId;
        _this.state = initialValue;
        _this.subscribe(noop); // stores are always kept alive
        Object.defineProperty(_this, "currentDraft", {
            // keeps jest from dying on revoked proxies..
            enumerable: false,
            configurable: true,
            writable: true,
            value: undefined
        });
        return _this;
    }
    Store.prototype.update = function (updater) {
        var _this = this;
        if (this.isRunningUpdate) {
            // update call from an update call (for example caused by merge)
            // just reuse the current draft
            if (!isFn(updater))
                fail$1("Illegal state");
            else
                avoidReturns(this.currentDraft, updater(this.currentDraft));
            return;
        }
        try {
            var baseState = this.state;
            this.state = produce(this.state, function (draft) {
                _this.isRunningUpdate = true;
                _this.currentDraft = draft;
                return normalizeUpdater(updater)(draft);
            });
            if (this.state !== baseState) {
                // skip default implementation
                var derivations = this.derivations.slice(1); // first derivation is the keep-alive noop, fine to skip
                derivations.forEach(function (d) { return d.propagateChanged(); });
                derivations.forEach(function (d) { return d.propagateReady(true); });
            }
        }
        finally {
            this.currentDraft = undefined;
            this.isRunningUpdate = false;
        }
    };
    Store.prototype.recompute = function () {
        return this.state;
    };
    Store.prototype.resume = function () {
        // happens only once
    };
    Store.prototype.suspend = function () {
        fail$1("nope");
    };
    Store.prototype.getCacheKey = function () {
        fail$1("nope");
    };
    Store.prototype.describe = function () {
        return this.name;
    };
    return Store;
}(BaseLens));
function createStore(initialValue, options) {
    return new Store(initialValue, options);
}

function normalizeUpdater(updater) {
    if (updater === undefined)
        return function () { return nothing; };
    if (isFn(updater))
        return updater;
    if (isObject(updater))
        return function (draft) { return isObject(draft) ? void Object.assign(draft, updater) : updater; };
    return function () { return updater; };
}
function avoidReturns(base, value) {
    if (value !== undefined && value !== base)
        return fail$1("This updater function should modify the current draft, and cannot return a complety new state");
}

var mergeCache = new Map();
var Merge = /** @class */ (function (_super) {
    __extends(Merge, _super);
    function Merge(bases) {
        var _this = _super.call(this) || this;
        _this.bases = bases;
        return _this;
        // TODO: check args
    }
    Merge.prototype.recompute = function () {
        // note, one of the deps has changed, so the output of the merge
        // is guaranteed to be different as well
        return this.bases.map(grabValue);
    };
    Merge.prototype.update = function (updater) {
        var _this = this;
        var drafts = [];
        var collectedResults;
        var grabNextDraft = function () {
            // probably optimizable
            _this.bases[drafts.length].update(function (draft) {
                drafts.push(draft);
                if (drafts.length < _this.bases.length)
                    grabNextDraft();
                else {
                    // we are now in the context of all producers of all bases
                    // let's call the updater function with those drafts!
                    var r = normalizeUpdater(updater)(drafts);
                    if (r === undefined)
                        collectedResults = drafts.slice();
                    else if (!Array.isArray(r) || r.length !== _this.bases.length)
                        return fail$1("updater for merge should return an array of length " + _this.bases.length + ", got: '" + JSON.stringify(r) + "'");
                    else
                        collectedResults = r;
                }
                var returnValue = collectedResults.pop();
                return returnValue === undefined ? nothing : returnValue;
            });
        };
        grabNextDraft();
    };
    Merge.prototype.resume = function () {
        var _this = this;
        mergeCache.set(getMergeCacheKey(this.bases), this);
        this.bases.forEach(function (b) { return b.registerDerivation(_this); });
    };
    Merge.prototype.suspend = function () {
        var _this = this;
        var ck = getMergeCacheKey(this.bases);
        if (mergeCache.get(ck) === this)
            mergeCache.delete(ck);
        this.bases.forEach(function (b) { return b.removeDerivation(_this); });
    };
    Merge.prototype.getCacheKey = function () {
        return undefined; // no simple cache key, but, see merge cache!
    };
    Merge.prototype.describe = function () {
        return ("merge(\n" + this.bases.map(function (b) { return b.describe(); }).join(",\n") + "\n)");
    };
    return Merge;
}(BaseLens));
function getMergeCacheKey(bases) {
    return bases.map(function (b) { return b.lensId; }).join(";");
}
function merge() {
    var lenses = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        lenses[_i] = arguments[_i];
    }
    var existing = mergeCache.get(getMergeCacheKey(lenses));
    if (existing)
        return existing;
    return new Merge(lenses);
}

var emptyArray = [];
Object.freeze(emptyArray);
function noop() { }
function fail$1(msg) {
    throw new Error("[remmi] " + msg);
}
function grabValue(lens) {
    return lens.value();
}
function _shallowEqual(ar1, ar2) {
    // TODO: optimize
    // TODO: use some library for this
    if (ar1 === ar2)
        return true;
    if (!ar1 || !ar2)
        return false;
    if (typeof ar1 === "object" && typeof ar2 === "object") {
        if (Array.isArray(ar1) !== Array.isArray(ar2))
            return false;
        if (Array.isArray(ar1)) {
            if (!Array.isArray(ar2))
                return false;
            if (ar1.length !== ar2.length)
                return false;
            for (var i = 0; i < ar1.length; i++)
                if (ar1[i] !== ar2[i])
                    return false;
            return true;
        }
        var aKeys = Object.keys(ar1).sort();
        var bKeys = Object.keys(ar2).sort();
        if (aKeys.length !== bKeys.length)
            return false;
        for (var index in aKeys) {
            var key = aKeys[index];
            if (key !== bKeys[index] || ar1[key] !== ar2[key])
                return false;
        }
        return true;
    }
    return false;
}
function shallowDiff(baseValue, newValue) {
    var diff = {
        changed: [],
        removed: [],
        added: []
    };
    // Optimize! For example, if base or new is empty
    var baseKeys = new Set(Object.keys(baseValue));
    var newKeys = new Set(Object.keys(newValue));
    baseKeys.forEach(function (key) {
        var base = baseValue[key];
        if (newKeys.has(key)) {
            if (newValue[key] !== base)
                diff.changed.push([key, newValue[key], base]);
        }
        else {
            diff.removed.push([key, base]);
        }
    });
    newKeys.forEach(function (key) {
        if (!baseKeys.has(key))
            diff.added.push([key, newValue[key]]);
    });
    return diff;
}
function each$1(value, cb) {
    if (Array.isArray(value)) {
        for (var i = 0; i < value.length; i++)
            cb(i, value[i]);
    }
    else {
        for (var key in value)
            cb(key, value[key]);
    }
}

function isFn(thing) {
    return typeof thing === "function";
}

function isObject(thing) {
    return (typeof thing === "object" &&
        thing !== null &&
        (thing.constructor === Object || thing.constructor === undefined));
}
function stringifyFunction(fn) {
    if (fn.name)
        return fn.name;
    var lines = fn.toString().split("\n");
    if (lines.length <= 4)
        return lines.map(trim).join(" ");
    return "<function>";
}
function trim(str) {
    return str.replace(/^\s+|\s+$/g, "");
}

var readListener;
function notifyRead(lens) {
    if (readListener)
        readListener(lens); // optimize
}
var Tracker = /** @class */ (function () {
    function Tracker(onInvalidate) {
        this.onInvalidate = onInvalidate;
        this.merge = new Merge([]);
        this.disposeMerge = this.merge.subscribe(noop);
    }
    Tracker.prototype.track = function (fn) {
        var dependencies = new Set();
        var prevListener = readListener;
        readListener = dependencies.add.bind(dependencies);
        try {
            return fn();
        }
        finally {
            var newDeps = Array.from(dependencies);
            if (!_shallowEqual(newDeps, this.getDependencies())) {
                // don't create merge if only one dep
                var nextMerge = newDeps.length === 0
                    ? (console.warn("[immer] The Tracker did not use any lenses"), createStore(undefined))
                    : newDeps.length === 1 ? newDeps[0] : merge.apply(void 0, newDeps);
                if (nextMerge !== this.merge) {
                    var disposeMerge = this.disposeMerge;
                    this.merge = nextMerge;
                    this.disposeMerge = this.merge.subscribe(this.onInvalidate);
                    disposeMerge();
                }
            }
            readListener = prevListener;
        }
    };
    Tracker.prototype.dispose = function () {
        this.disposeMerge();
    };
    Tracker.prototype.getDependencies = function () {
        return this.merge instanceof Merge ? this.merge.bases : [this.merge];
    };
    return Tracker;
}());
function autorun(fn) {
    var t = new Tracker(function () {
        t.track(fn);
    });
    var res = t.dispose.bind(t);
    t.track(fn);
    return res;
}

function grabIndex(_v, idx) {
    return idx;
}
function keySelector(value) {
    if (Array.isArray(value))
        return value.map(grabIndex);
    if (value !== null && typeof value === "object")
        return Object.keys(value);
    return emptyArray;
}
function keys(lens) {
    return lens.do(select(keySelector), shallowEqual$1);
}

function fork(recordActions) {
    if (recordActions === void 0) { recordActions = false; }
    return function (lens) {
        var fork = createStore(lens.value());
        if (!recordActions)
            return fork;
        var recordedUpdates = [];
        var recording = true;
        return Object.assign(fork.transform({
            cacheKey: undefined,
            description: "recorder(true)",
            onUpdate: function (updater, next) {
                if (recording)
                    recordedUpdates.push(updater);
                next(updater);
            }
        }), {
            pause: function () {
                recording = false;
            },
            resume: function () {
                recording = true;
            },
            reset: function () {
                recordedUpdates.splice(0);
            },
            /**
             * Replays the recorded update onto the target
             * This process is atomic; if one of the updaters throws an exception,
             * the target will be unmodified
             */
            replay: function (target) {
                if (target === void 0) { target = this; }
                target.update(function (draft) {
                    recordedUpdates.forEach(function (u) {
                        
                        u(draft);
                    });
                });
            }
        });
    };
}

var All = { All: true };
function all(lens) {
    // TODO: express all in terms of mapReduce?
    // all sublenses are always cached and reconciled until all is GC-ed itself
    var subLensCache = new Map();
    return lens.do(keys).transform({
        cacheKey: All,
        onNext: function (nextValue) {
            // source.keys() already includes shallow comparision, so
            // base value has always introduced or removed entries here
            var newCache = new Map();
            var lenses = nextValue.map(function (key) {
                var subLens = subLensCache.get(key) || lens.select(key);
                newCache.set(key, subLens);
                return subLens;
            });
            subLensCache = newCache;
            return lenses;
        },
        onUpdate: function () {
            // question: or make this actually possible, and just cal on base?
            fail("Cannot call update on `.all()`, call update on an individual lens instead");
        },
        description: "all()"
    });
}

// TODO: make index a number for arrays?
function mapReduce(mapper, reducer) {
    return function (baseLens) {
        // all sublenses are always cached and reconciled until all is GC-ed itself
        var entries = new Map();
        var prevBaseValue;
        return baseLens.transform({
            cacheKey: undefined,
            onNext: function (sourceValue, previousResult, self) {
                if (!sourceValue || typeof sourceValue !== "object")
                    return sourceValue; // map Reduce doesn't do anything for primitive types
                // shallowly-changed?
                var _a = shallowDiff(
                // if not hot, we should do a full diff..
                (self.hot && prevBaseValue) || (Array.isArray(sourceValue) ? [] : {}), sourceValue), changed = _a.changed, added = _a.added, removed = _a.removed;
                prevBaseValue = sourceValue;
                // short-circuit
                if (changed.length === 0 &&
                    added.length === 0 &&
                    removed.length === 0)
                    return previousResult;
                var changes = {
                    added: [],
                    removed: [],
                    updated: []
                };
                // new entries
                added.forEach(function (_a) {
                    var key = _a[0], value = _a[1];
                    var rootLens = createStore(value);
                    var mappedLens = mapper(rootLens, key);
                    var disposer = mappedLens.subscribe(noop);
                    entries.set(key, { rootLens: rootLens, mappedLens: mappedLens, disposer: disposer });
                    changes.added.push([key, mappedLens.value()]);
                });
                // removed entries
                removed.forEach(function (_a) {
                    var key = _a[0];
                    var entry = entries.get(key);
                    changes.removed.push([key, entry.mappedLens.value()]);
                    entries.delete(key);
                });
                // changed entries
                changed.forEach(function (_a) {
                    var key = _a[0], newBaseValue = _a[1];
                    var entry = entries.get(key);
                    var oldMappedValue = entry.mappedLens.value();
                    entry.rootLens.update(newBaseValue);
                    var newMappedValue = entry.mappedLens.value();
                    if (oldMappedValue !== newMappedValue)
                        changes.updated.push([key, newMappedValue, oldMappedValue]);
                });
                // short-circuit
                if (changed.length === 0 &&
                    added.length === 0 &&
                    removed.length === 0)
                    return previousResult;
                // run reducer
                return reducer(previousResult, changes, sourceValue);
            },
            onUpdate: function () {
                // question: or make this actually possible, and just cal on base?
                fail$1("Cannot call update on `.mapReduce()`, call update on an individual lens instead");
            },
            description: "mapReduce()" // TODO: improve
        });
    };
}

function model(modelDefinition) {
    return function (lens) {
        var modelLens = lens.transform({
            cacheKey: modelDefinition,
            description: "model(" + modelDefinition.name + ")"
        });
        if (modelLens.modelDefinition === modelDefinition)
            return modelLens;
        var memberDefinitions = modelDefinition(lens);
        // optimize: only check in production
        // hide baseLens internally to avoid naming conflicts on members?
        for (var key in memberDefinitions)
            if (key in modelLens)
                fail$1("Cannot introduce model property '" + key + "', it would override a core lens property");
        Object.assign(modelLens, memberDefinitions, { modelDefinition: modelDefinition });
        return modelLens;
    };
}

var ReadOnly = { ReadOnly: true };
function readOnly(lens) {
    return lens.transform({
        cacheKey: ReadOnly,
        onUpdate: function () {
            fail$1("Read only lens");
        },
        description: "readOnly()"
    });
}

function select(selector) {
    return function (lens) {
        if (typeof selector === "function")
            return selectFn(lens, selector);
        if (typeof selector === "string" || typeof selector === "number")
            return selectProp(lens, "" + selector);
        return fail$1("Invalid selector, expected number, string or function. Got: " +
            typeof selector);
    };
}
function selectFn(lens, selector) {
    return lens.transform({
        cacheKey: selector,
        onNext: function (newBaseValue) {
            return selector(newBaseValue);
        },
        onUpdate: function (updater, next) {
            next(function (draft) {
                var baseState = selector(draft);
                avoidReturns(baseState, updater(baseState));
            });
        },
        description: "select(" + stringifyFunction(selector) + ")"
    });
}
function selectProp(lens, key) {
    return Object.assign(lens.transform({
        cacheKey: key,
        onNext: function (newBaseValue) {
            if (newBaseValue === null || newBaseValue === undefined)
                return undefined;
            if (typeof newBaseValue === "object")
                return newBaseValue[key];
            return undefined;
        },
        onUpdate: function (updater, next) {
            next(function (draft) {
                var baseState = draft[key];
                var res = updater(baseState);
                if (res === nothing)
                    draft[key] = undefined;
                else if (res !== undefined)
                    draft[key] = res;
            });
        },
        description: "\"" + key + "\""
    }), { key: key });
}

var ShallowEqual = { ShallowEqualLens: true };
function shallowEqual$1(lens) {
    return lens.transform({
        cacheKey: ShallowEqual,
        onNext: function (newBaseValue, currentValue) {
            if (_shallowEqual(currentValue, newBaseValue))
                return currentValue;
            return newBaseValue;
        },
        description: "shallowEqual()"
    });
}

function defaultLog(newValue, _oldValue, lens) {
    console.log("[remmi] " + lens + ": " + JSON.stringify(newValue));
}
function tap(logger) {
    if (logger === void 0) { logger = defaultLog; }
    return function (lens) {
        return lens.transform({
            cacheKey: logger,
            onNext: function (newBaseValue, currentValue, self) {
                logger(newBaseValue, currentValue, self.toString());
                return newBaseValue;
            },
            description: "log()"
        });
    };
}

var AutoRender = /** @class */ (function (_super) {
    __extends(AutoRender, _super);
    function AutoRender() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    AutoRender.prototype.render = function () {
        var _this = this;
        if (!this.tracker) {
            this.tracker = new Tracker(function () {
                _this.forceUpdate();
            });
        }
        return this.tracker.track(this.props.children);
    };
    AutoRender.prototype.componentWillUnmount = function () {
        if (this.tracker)
            this.tracker.dispose();
    };
    return AutoRender;
}(React.PureComponent));
var RenderLens = /** @class */ (function (_super) {
    __extends(RenderLens, _super);
    function RenderLens() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.onChange = function () {
            _this.forceUpdate();
        };
        return _this;
    }
    RenderLens.prototype.render = function () {
        return this.props.renderer(this.props.lens.value(), this.props.lens);
    };
    RenderLens.prototype.componentDidMount = function () {
        this.subscription = this.props.lens.subscribe(this.onChange);
    };
    RenderLens.prototype.componentDidUpdate = function (prevProps) {
        // lens changed
        if (prevProps.lens !== this.props.lens) {
            var prevSub = this.subscription;
            this.componentDidMount();
            prevSub(); // clear previous listener after attaching new one
        }
    };
    RenderLens.prototype.componentWillUnmount = function () {
        this.subscription();
    };
    return RenderLens;
}(React.PureComponent));
var RenderLenses = /** @class */ (function (_super) {
    __extends(RenderLenses, _super);
    function RenderLenses() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.itemRenderer = function (lens) {
            return (React.createElement(RenderLens, { key: "" + lens.key, lens: lens, renderer: _this.props.renderer }));
        };
        _this.allRenderer = function (lenses) {
            return React.createElement(React.Fragment, null, lenses.map(_this.itemRenderer));
        };
        return _this;
    }
    RenderLenses.prototype.render = function () {
        return (React.createElement(RenderLens, { lens: this.props.lens.do(all), renderer: this.allRenderer }));
    };
    return RenderLenses;
}(React.PureComponent));
function autoRender(fn) {
    return React.createElement(AutoRender, null, fn);
}
function render(renderer) {
    return function (lens) { return React.createElement(RenderLens, { lens: lens, renderer: renderer }); };
}
function renderAll(renderer) {
    return function (lens) { return React.createElement(RenderLenses, { lens: lens, renderer: renderer }); };
}

function invariant(predicate) {
    return function (lens) {
        return lens.subscribe(function (value) {
            var res = predicate(value);
            if (res === false)
                return fail$1("Invariant failed: '" + stringifyFunction(predicate) + "'");
            if (typeof res === "string" && res)
                throw new Error(res);
            return;
        });
    };
}

var Subscription = /** @class */ (function (_super) {
    __extends(Subscription, _super);
    function Subscription(base, handler) {
        var _this = _super.call(this) || this;
        _this.base = base;
        _this.handler = handler;
        _this.disposed = false;
        _this.dispose = function () {
            if (!_this.disposed)
                _this.base.removeDerivation(_this);
            _this.disposed = true;
        };
        base.registerDerivation(_this);
        return _this;
    }
    Subscription.prototype.recompute = function () {
        if (this.disposed)
            fail("Illegal state");
        this.handler(this.base.value());
        return undefined;
    };
    Subscription.prototype.resume = function () {
        fail("Illegal state");
    };
    Subscription.prototype.suspend = function () {
        fail("Illegal state");
    };
    Subscription.prototype.update = function () {
        fail("Illegal state");
    };
    Subscription.prototype.getCacheKey = function () {
        return undefined;
    };
    Subscription.prototype.describe = function () {
        return "subscribe(" + stringifyFunction(this.handler) + ")";
    };
    return Subscription;
}(BaseLens));
function subscribe(subscription) {
    return function (lens) {
        return new Subscription(lens, subscription).dispose;
    };
}

function toStream(lens) {
    var _a;
    return _a = {
            subscribe: function (observer) {
                var unsubscribe = lens.subscribe(typeof observer === "function" ? observer : observer.next.bind(observer));
                return {
                    unsubscribe: unsubscribe
                };
            }
        }, _a[observableSymbol()] = function () {
            return this;
        }, _a;
}

function fromStream(stream) {
    return function (lens) {
        var subscription = stream.subscribe(lens.update.bind(lens));
        return function () { return subscription.unsubscribe(); };
    };
}

function connect(handler) {
    return function (lens) {
        var subscriptionDisposer;
        var isHandlingIncomingUpdate = false;
        var isHandlingSubscription = false;
        var handlerDisposer = handler(function (callback) {
            if (subscriptionDisposer)
                return fail$1("connect: subscribeToStore was called already");
            subscriptionDisposer = lens.subscribe(function (v) {
                if (!isHandlingIncomingUpdate) {
                    isHandlingSubscription = true;
                    try {
                        callback(v); // avoid loopback!
                    }
                    finally {
                        isHandlingSubscription = false;
                    }
                }
            });
            return subscriptionDisposer;
        }, function (updater) {
            if (isHandlingSubscription)
                return;
            isHandlingIncomingUpdate = true;
            try {
                lens.update(updater);
            }
            finally {
                isHandlingIncomingUpdate = false;
            }
        });
        return function () {
            subscriptionDisposer && subscriptionDisposer();
            handlerDisposer && handlerDisposer();
        };
    };
}

function map(mapper) {
    return mapReduce(function (lens, index) { return lens.do(select(function (value) { return mapper(value, index); })); }, function (prev, changes, source) {
        if (prev === undefined)
            prev = Array.isArray(source) ? [] : {};
        var isArray = Array.isArray(prev);
        var res = Array.isArray(prev) ? prev.slice() : __assign({}, prev);
        changes.updated.forEach(function (_a) {
            var key = _a[0], value = _a[1];
            res[key] = value;
        });
        if (changes.removed.length) {
            if (isArray)
                res.length -= changes.removed.length;
            else
                changes.removed.forEach(function (_a) {
                    var key = _a[0];
                    delete res[key];
                });
        }
        else
            changes.added.forEach(function (_a) {
                var key = _a[0], value = _a[1];
                if (isArray)
                    res.push(value);
                else
                    res[key] = value;
            });
        return res;
    });
}

function filter(predicate) {
    return function (lens) { return lens.do(map(function (value, index) { return [predicate(value, index), value]; }), select(function (mapped) {
        var isArray = Array.isArray(mapped);
        var res = isArray ? [] : {};
        each$1(mapped, function (index, value) {
            if (value[0] === true)
                if (isArray)
                    res.push(value[1]);
                else
                    res[index] = value[1];
        });
        return res;
    }), shallowEqual$1 // Optimize: inefficient, should, during map, compare with previous result
    // TODO: Idea: pass known previous state as second argument to select funtions
    // Also, shallowEqual selector can be implemented using `select` that way
    ); };
}

exports.all = all;
exports.autoRender = autoRender;
exports.autorun = autorun;
exports.connect = connect;
exports.createStore = createStore;
exports.filter = filter;
exports.fork = fork;
exports.fromStream = fromStream;
exports.invariant = invariant;
exports.isLens = isLens;
exports.keys = keys;
exports.map = map;
exports.mapReduce = mapReduce;
exports.merge = merge;
exports.model = model;
exports.readOnly = readOnly;
exports.render = render;
exports.renderAll = renderAll;
exports.select = select;
exports.shallowEqual = shallowEqual$1;
exports.subscribe = subscribe;
exports.tap = tap;
exports.toStream = toStream;
//# sourceMappingURL=remmi.js.map
