import { Disposer, Handler, Transformer } from "../internal";
export declare type IConnectionHandler<T = any> = (subscribeToStore: (callback: Handler<T>) => Disposer, writeToStore: (value: T) => void) => (void | Disposer);
/**
 * connect can be used to setup a uni- or bidirectional connection with arbitrarily remote sources.
 *
 * When setting up a bi-directional connection, it makes sure that subscriptions are not superflously called during an update,
 * and that update is not superfluously called during a subscription. (For asynchronous processes
 * this might not be sufficient, in which case one might to keep some adminstration for that in the handler).
 *
 * @example
 * // This set's up a bidirectional sink between two remmi lenses
 * let disposer = store1.do(connect((subscribe, sink) => {
 *     subscribe(v => {
 *         store2.update(v)
 *     })
 *     return store2.subscribe(v => sink(v))
 * }))
 *
 * @example
 * // Find the stocks store and update it automatically from a websocket
 * const store$ = createStore({
 *   stocks: {}
 * });
 * store$.do(
 *   select("stocks"),
 *   connect((_, sink) => {
 *     // See: https://iextrading.com/developer/docs/#websockets
 *     const socket = socketIo("https://ws-api.iextrading.com/1.0/tops");
 *     socket.on("connect", () => {
 *       socket.emit("subscribe", "snap,fb,aig+,amd,amzn,ge,nke");
 *     });
 *     socket.on("message", data => {
 *       const message = JSON.parse(data);
 *       sink(stocks => {
 *         stocks[message.symbol] = message;
 *       });
 *     });
 *   })
 * );
 */
export declare function connect<T = any>(handler: IConnectionHandler<T>): Transformer<T, Disposer | undefined>;
